@@ -28,6 +28,9 @@
 // operators, and so this lint is triggered unnecessarily.
 #![allow(clippy::suspicious_arithmetic_impl)]
 
+#[cfg(feature = "alloc")]
+extern crate alloc;
+
 #[cfg(test)]
 #[macro_use]
 extern crate std;
@@ -37,7 +40,7 @@ use core::borrow::Borrow;
 use core::fmt;
 use core::iter::Sum;
 use core::ops::{Add, AddAssign, Mul, MulAssign, Neg, Sub, SubAssign};
-use ff::Field;
+use ff::{BatchInverter, Field};
 use group::{
     cofactor::{CofactorCurve, CofactorCurveAffine, CofactorGroup},
     prime::PrimeGroup,
@@ -46,6 +49,9 @@ use group::{
 use rand_core::RngCore;
 use subtle::{Choice, ConditionallySelectable, ConstantTimeEq, CtOption};
 
+#[cfg(feature = "alloc")]
+use alloc::vec::Vec;
+
 #[cfg(feature = "alloc")]
 use group::WnafGroup;
 
@@ -56,7 +62,12 @@ mod fr;
 pub use bls12_381::Scalar as Fq;
 pub use fr::Fr;
 
-/// A better name than Fr.
+/// Represents an element of the base field $\mathbb{F}_q$ of the Jubjub elliptic curve
+/// construction.
+pub type Base = Fq;
+
+/// Represents an element of the scalar field $\mathbb{F}_r$ of the Jubjub elliptic curve
+/// construction.
 pub type Scalar = Fr;
 
 const FR_MODULUS_BYTES: [u8; 32] = [
@@ -522,6 +533,99 @@ impl AffinePoint {
         })
     }
 
+    /// Attempts to interpret a batch of byte representations of affine points.
+    ///
+    /// Returns None for each element if it is not on the curve, or is non-canonical
+    /// according to ZIP 216.
+    #[cfg(feature = "alloc")]
+    pub fn batch_from_bytes(items: impl Iterator<Item = [u8; 32]>) -> Vec<CtOption<Self>> {
+        use ff::BatchInvert;
+
+        #[derive(Clone, Copy, Default)]
+        struct Item {
+            sign: u8,
+            v: Fq,
+            numerator: Fq,
+            denominator: Fq,
+        }
+
+        impl ConditionallySelectable for Item {
+            fn conditional_select(a: &Self, b: &Self, choice: Choice) -> Self {
+                Item {
+                    sign: u8::conditional_select(&a.sign, &b.sign, choice),
+                    v: Fq::conditional_select(&a.v, &b.v, choice),
+                    numerator: Fq::conditional_select(&a.numerator, &b.numerator, choice),
+                    denominator: Fq::conditional_select(&a.denominator, &b.denominator, choice),
+                }
+            }
+        }
+
+        let items: Vec<_> = items
+            .map(|mut b| {
+                // Grab the sign bit from the representation
+                let sign = b[31] >> 7;
+
+                // Mask away the sign bit
+                b[31] &= 0b0111_1111;
+
+                // Interpret what remains as the v-coordinate
+                Fq::from_bytes(&b).map(|v| {
+                    // -u^2 + v^2 = 1 + d.u^2.v^2
+                    // -u^2 = 1 + d.u^2.v^2 - v^2    (rearrange)
+                    // -u^2 - d.u^2.v^2 = 1 - v^2    (rearrange)
+                    // u^2 + d.u^2.v^2 = v^2 - 1     (flip signs)
+                    // u^2 (1 + d.v^2) = v^2 - 1     (factor)
+                    // u^2 = (v^2 - 1) / (1 + d.v^2) (isolate u^2)
+                    // We know that (1 + d.v^2) is nonzero for all v:
+                    //   (1 + d.v^2) = 0
+                    //   d.v^2 = -1
+                    //   v^2 = -(1 / d)   No solutions, as -(1 / d) is not a square
+
+                    let v2 = v.square();
+
+                    Item {
+                        v,
+                        sign,
+                        numerator: (v2 - Fq::one()),
+                        denominator: Fq::one() + EDWARDS_D * v2,
+                    }
+                })
+            })
+            .collect();
+
+        let mut denominators: Vec<_> = items
+            .iter()
+            .map(|item| item.map(|item| item.denominator).unwrap_or(Fq::zero()))
+            .collect();
+        denominators.iter_mut().batch_invert();
+
+        items
+            .into_iter()
+            .zip(denominators.into_iter())
+            .map(|(item, inv_denominator)| {
+                item.and_then(
+                    |Item {
+                         v, sign, numerator, ..
+                     }| {
+                        (numerator * inv_denominator).sqrt().and_then(|u| {
+                            // Fix the sign of `u` if necessary
+                            let flip_sign = Choice::from((u.to_bytes()[0] ^ sign) & 1);
+                            let u_negated = -u;
+                            let final_u = Fq::conditional_select(&u, &u_negated, flip_sign);
+
+                            // If u == 0, flip_sign == sign_bit. We therefore want to reject the
+                            // encoding as non-canonical if all of the following occur:
+                            // - u == 0
+                            // - flip_sign == true
+                            let u_is_zero = u.ct_eq(&Fq::zero());
+                            CtOption::new(AffinePoint { u: final_u, v }, !(u_is_zero & flip_sign))
+                        })
+                    },
+                )
+            })
+            .collect()
+    }
+
     /// Returns the `u`-coordinate of this point.
     pub fn get_u(&self) -> Fq {
         self.u
@@ -736,23 +840,16 @@ impl ExtendedPoint {
     fn batch_normalize(p: &[Self], q: &mut [AffinePoint]) {
         assert_eq!(p.len(), q.len());
 
-        let mut acc = Fq::one();
         for (p, q) in p.iter().zip(q.iter_mut()) {
-            // We use the `u` field of `AffinePoint` to store the product
-            // of previous z-coordinates seen.
-            q.u = acc;
-            acc *= &p.z;
+            // We use the `u` field of `AffinePoint` to store the z-coordinate being
+            // inverted, and the `v` field for scratch space.
+            q.u = p.z;
         }
 
-        // This is the inverse, as all z-coordinates are nonzero.
-        acc = acc.invert().unwrap();
+        BatchInverter::invert_with_internal_scratch(q, |q| &mut q.u, |q| &mut q.v);
 
         for (p, q) in p.iter().zip(q.iter_mut()).rev() {
-            // Compute tmp = 1/z
-            let tmp = q.u * acc;
-
-            // Cancel out z-coordinate in denominator of `acc`
-            acc *= &p.z;
+            let tmp = q.u;
 
             // Set the coordinates to the correct value
             q.u = p.u * &tmp; // Multiply by 1/z
@@ -985,25 +1082,12 @@ impl Default for ExtendedPoint {
 ///
 /// This costs 5 multiplications per element, and a field inversion.
 pub fn batch_normalize<'a>(v: &'a mut [ExtendedPoint]) -> impl Iterator<Item = AffinePoint> + 'a {
-    let mut acc = Fq::one();
+    // We use the `t1` field of `ExtendedPoint` for scratch space.
+    BatchInverter::invert_with_internal_scratch(v, |p| &mut p.z, |p| &mut p.t1);
+
     for p in v.iter_mut() {
-        // We use the `t1` field of `ExtendedPoint` to store the product
-        // of previous z-coordinates seen.
-        p.t1 = acc;
-        acc *= &p.z;
-    }
-
-    // This is the inverse, as all z-coordinates are nonzero.
-    acc = acc.invert().unwrap();
-
-    for p in v.iter_mut().rev() {
         let mut q = *p;
-
-        // Compute tmp = 1/z
-        let tmp = q.t1 * acc;
-
-        // Cancel out z-coordinate in denominator of `acc`
-        acc *= &q.z;
+        let tmp = q.z;
 
         // Set the coordinates to the correct value
         q.u *= &tmp; // Multiply by 1/z
@@ -1471,6 +1555,11 @@ fn test_batch_normalize() {
     }
 
     let expected: std::vec::Vec<_> = v.iter().map(|p| AffinePoint::from(*p)).collect();
+    let mut result0 = vec![AffinePoint::identity(); v.len()];
+    ExtendedPoint::batch_normalize(&v, &mut result0);
+    for i in 0..10 {
+        assert!(expected[i] == result0[i]);
+    }
     let result1: std::vec::Vec<_> = batch_normalize(&mut v).collect();
     for i in 0..10 {
         assert!(expected[i] == result1[i]);
@@ -1786,12 +1875,15 @@ fn test_serialization_consistency() {
         ],
     ];
 
-    for expected_serialized in v {
+    let batched = AffinePoint::batch_from_bytes(v.iter().cloned());
+
+    for (expected_serialized, batch_deserialized) in v.into_iter().zip(batched.into_iter()) {
         assert!(p.is_on_curve_vartime());
         let affine = AffinePoint::from(p);
         let serialized = affine.to_bytes();
         let deserialized = AffinePoint::from_bytes(serialized).unwrap();
         assert_eq!(affine, deserialized);
+        assert_eq!(affine, batch_deserialized.unwrap());
         assert_eq!(expected_serialized, serialized);
         p += gen;
     }
